@@ -18,15 +18,16 @@
 
 // The autoconf system provides us with the NO_IPINFO define. 
 // Littering the code with #ifndef NO_IPINFO (double negative)
-// does not benefit readabilty. So here we invert the sense of the
+// does not benefit readability. So here we invert the sense of the
 // define. 
 //
-// Similarly, this include file should be included unconditially. 
+// Similarly, this include file should be included unconditionally. 
 // It will evaluate to nothing if we don't need it. 
 
 #ifndef NO_IPINFO
-#define IPINFO
 
+#ifndef IPINFO
+#define IPINFO
 
 extern int ipinfo_no;
 extern int ipinfo_max;
@@ -38,4 +39,5 @@ char *fmt_ipinfo(ip_t *addr);
 int get_iiwidth(void);
 int is_printii(void);
 
-#endif
+#endif /* IPINFO */
+#endif /* NO_IPINFO */
