@@ -16,19 +16,21 @@
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
+#include "config.h"
+
 #include <sys/types.h>
-#include <config.h>
-#include <netdb.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <netinet/in.h>
-#include <sys/socket.h>
 #include <unistd.h>
-#include <strings.h>
-#include <time.h>
 #include <errno.h>
 #include <string.h>
+#include <strings.h>
+
+#include <netdb.h>
+#include <netinet/in.h>
+#include <sys/socket.h>
+#include <time.h>
 #include <ctype.h>
 #include <assert.h>
 #include <fcntl.h>
@@ -41,9 +43,7 @@
 #include "dns.h"
 #include "report.h"
 #include "net.h"
-#ifndef NO_IPINFO
 #include "asn.h"
-#endif
 #include "version.h"
 
 
@@ -67,6 +67,7 @@ int   PrintHelp = 0;
 int   MaxPing = 10;
 int   ForceMaxPing = 0;
 float WaitTime = 1.0;
+float GraceTime = 5.0;
 char *Hostname = NULL;
 char *InterfaceAddress = NULL;
 char  LocalHostname[128];
@@ -76,6 +77,9 @@ int   enablempls = 0;
 int   cpacketsize = 64;          /* default packet size */
 int   bitpattern = 0;
 int   tos = 0;
+#ifdef SO_MARK
+int   mark = -1;
+#endif
 int   reportwide = 0;
 int af = DEFAULT_AF;
 int mtrtype = IPPROTO_ICMP;     /* Use ICMP as default packet type */
@@ -85,8 +89,11 @@ int  fstTTL = 1;                /* default start at first hop */
 /*int maxTTL = MaxHost-1;  */     /* max you can go is 255 hops */
 int   maxTTL = 30;              /* inline with traceroute */
                                 /* end ttl window stuff. */
-int remoteport = 80;            /* for TCP tracing */
-int timeout = 10 * 1000000;     /* for TCP tracing */
+int maxUnknown = 5;				/* stop send package */
+                                /*when larger than this count */
+int remoteport = 0;            /* for TCP tracing */
+int localport = 0;             /* for UDP tracing */
+int tcp_timeout = 10 * 1000000;     /* for TCP tracing */
 
 
 /* default display field(defined by key in net.h) and order */
@@ -254,53 +261,70 @@ void parse_arg (int argc, char **argv)
 {
   int opt;
   int i;
+  /* IMPORTANT: when adding or modifying an option:
+       0/ try to find a somewhat logical order;
+       1/ add the long option name in "long_options" below;
+       2/ add the short option name in the "getopt_long" call;
+       3/ update the man page (use the same order);
+       4/ update the help message (see PrintHelp).
+   */
   static struct option long_options[] = {
-    { "version", 0, 0, 'v' },
-    { "help", 0, 0, 'h' },
+    /* option name, has argument, NULL, short name */
+    { "help",           0, NULL, 'h' },
+    { "version",        0, NULL, 'v' },
 
-    { "report", 0, 0, 'r' },
-    { "report-wide", 0, 0, 'w' },
-    { "xml", 0, 0, 'x' },
-    { "curses", 0, 0, 't' },
-    { "gtk", 0, 0, 'g' },
-    { "raw", 0, 0, 'l' },
-    { "csv", 0, 0, 'C' },
-    { "split", 0, 0, 'p' },     /* BL */
-    				/* maybe above should change to -d 'x' */
+    { "inet",           0, NULL, '4' }, /* IPv4 only */
+    { "inet6",          0, NULL, '6' }, /* IPv6 only */
 
-    { "order", 1, 0, 'o' },	/* fileds to display & their order */
+    { "filename",       1, NULL, 'F' },
 
-    { "interval", 1, 0, 'i' },
-    { "report-cycles", 1, 0, 'c' },
-    { "psize", 1, 0, 's' },	/* changed 'p' to 's' to match ping option
-				   overload psize<0, ->rand(min,max) */
-    { "bitpattern", 1, 0, 'B' },/* overload b>255, ->rand(0,255) */
-    { "tos", 1, 0, 'Q' },	/* typeof service (0,255) */
-    { "mpls", 0, 0, 'e' },
-    { "no-dns", 0, 0, 'n' },
-    { "show-ips", 0, 0, 'b' },
-    { "address", 1, 0, 'a' },
-    { "first-ttl", 1, 0, 'f' },	/* -f & -m are borrowed from traceroute */
-    { "filename", 1, 0, 'F' },
-    { "max-ttl", 1, 0, 'm' },
-    { "udp", 0, 0, 'u' },	/* UDP (default is ICMP) */
-    { "tcp", 0, 0, 'T' },	/* TCP (default is ICMP) */
-    { "port", 1, 0, 'P' },      /* target port number for TCP */
-    { "timeout", 1, 0, 'Z' },   /* timeout for TCP sockets */
-    { "inet", 0, 0, '4' },	/* IPv4 only */
-    { "inet6", 0, 0, '6' },	/* IPv6 only */
-#ifndef NO_IPINFO
-    { "ipinfo", 1, 0, 'y' },    /* IP info lookup */
-    { "aslookup", 0, 0, 'z' },  /* Do AS lookup (--ipinfo 0) */
+    { "report",         0, NULL, 'r' },
+    { "report-wide",    0, NULL, 'w' },
+    { "xml",            0, NULL, 'x' },
+    { "curses",         0, NULL, 't' },
+    { "gtk",            0, NULL, 'g' },
+    { "raw",            0, NULL, 'l' },
+    { "csv",            0, NULL, 'C' },
+    { "json",           0, NULL, 'j' },
+    { "displaymode",    1, NULL, 'd' },
+    { "split",          0, NULL, 'p' }, /* BL */
+                                        /* maybe above should change to -d 'x' */
+
+    { "no-dns",         0, NULL, 'n' },
+    { "show-ips",       0, NULL, 'b' },
+    { "order",          1, NULL, 'o' }, /* fields to display & their order */
+#ifdef IPINFO
+    { "ipinfo",         1, NULL, 'y' }, /* IP info lookup */
+    { "aslookup",       0, NULL, 'z' }, /* Do AS lookup (--ipinfo 0) */
+#endif
+
+    { "interval",       1, NULL, 'i' },
+    { "report-cycles",  1, NULL, 'c' },
+    { "psize",          1, NULL, 's' }, /* overload psize<0, ->rand(min,max) */
+    { "bitpattern",     1, NULL, 'B' }, /* overload B>255, ->rand(0,255) */
+    { "tos",            1, NULL, 'Q' }, /* typeof service (0,255) */
+    { "mpls",           0, NULL, 'e' },
+    { "address",        1, NULL, 'a' },
+    { "first-ttl",      1, NULL, 'f' }, /* -f & -m are borrowed from traceroute */
+    { "max-ttl",        1, NULL, 'm' },
+	{ "max-unknown",    1, NULL, 'U' },
+    { "udp",            0, NULL, 'u' }, /* UDP (default is ICMP) */
+    { "tcp",            0, NULL, 'T' }, /* TCP (default is ICMP) */
+    { "sctp",           0, NULL, 'S' }, /* SCTP (default is ICMP) */
+    { "port",           1, NULL, 'P' }, /* target port number for TCP/SCTP/UDP */
+    { "localport",      1, NULL, 'L' }, /* source port number for UDP */
+    { "timeout",        1, NULL, 'Z' }, /* timeout for TCP sockets */
+    { "gracetime",      1, NULL, 'G' }, /* graceperiod for replies after last probe */
+#ifdef SO_MARK
+    { "mark",           1, NULL, 'M' }, /* use SO_MARK */
 #endif
     { 0, 0, 0, 0 }
   };
 
   opt = 0;
   while(1) {
-    /* added f:m:o: byMin */
     opt = getopt_long(argc, argv,
-		      "vhrwxtglCpo:B:i:c:s:Q:ena:f:m:uTP:Zby:z46", long_options, NULL);
+		      "hv46F:rwxtglCjpnbo:y:zi:c:s:B:Q:ea:f:m:U:uTSP:L:Z:G:M:", long_options, NULL);
     if(opt == -1)
       break;
 
@@ -334,10 +358,16 @@ void parse_arg (int argc, char **argv)
     case 'C':
       DisplayMode = DisplayCSV;
       break;
+    case 'j':
+      DisplayMode = DisplayJSON;
+      break;
     case 'x':
       DisplayMode = DisplayXML;
       break;
 
+    case 'd':
+      display_mode = (atoi (optarg)) % 3;
+      break;
     case 'c':
       MaxPing = atoi (optarg);
       ForceMaxPing = 1;
@@ -389,6 +419,12 @@ void parse_arg (int argc, char **argv)
 	fstTTL = maxTTL;
       }
       break;
+	case 'U':
+		maxUnknown = atoi(optarg);
+		if (maxUnknown < 1) {
+			maxUnknown = 1;
+		}
+		break;
     case 'o':
       /* Check option before passing it on to fld_active. */
       if (strlen (optarg) > MAXFLD) {
@@ -408,6 +444,13 @@ void parse_arg (int argc, char **argv)
       if (bitpattern > 255)
 	bitpattern = -1;
       break;
+    case 'G':
+      GraceTime = atof (optarg);
+      if (GraceTime <= 0.0) {
+        fprintf (stderr, "mtr: wait time must be positive\n");
+        exit (1);
+      }
+      break;
     case 'Q':
       tos = atoi (optarg);
       if (tos > 255 || tos < 0) {
@@ -418,18 +461,30 @@ void parse_arg (int argc, char **argv)
       break;
     case 'u':
       if (mtrtype != IPPROTO_ICMP) {
-        fprintf(stderr, "-u and -T are mutually exclusive.\n");
+        fprintf(stderr, "-u , -T and -S are mutually exclusive.\n");
         exit(EXIT_FAILURE);
       }
       mtrtype = IPPROTO_UDP;
       break;
     case 'T':
       if (mtrtype != IPPROTO_ICMP) {
-        fprintf(stderr, "-u and -T are mutually exclusive.\n");
+        fprintf(stderr, "-u , -T and -S are mutually exclusive.\n");
         exit(EXIT_FAILURE);
       }
+      if (!remoteport) {
+        remoteport = 80;
+      }
       mtrtype = IPPROTO_TCP;
       break;
+    case 'S':
+      if (mtrtype != IPPROTO_ICMP) {
+        fprintf(stderr, "-u , -T and -S are mutually exclusive.\n");
+        exit(EXIT_FAILURE);
+      }
+      if (!remoteport) {
+        remoteport = 80;
+      }
+      mtrtype = IPPROTO_SCTP;
     case 'b':
       show_ips = 1;
       break;
@@ -440,9 +495,16 @@ void parse_arg (int argc, char **argv)
         exit(EXIT_FAILURE);
       }
       break;
+    case 'L':
+      localport = atoi(optarg);
+      if (localport > 65535 || localport < MinPort) {
+        fprintf(stderr, "Illegal local port number.\n");
+        exit(EXIT_FAILURE);
+      }
+      break;
     case 'Z':
-      timeout = atoi(optarg);
-      timeout *= 1000000;
+      tcp_timeout = atoi(optarg);
+      tcp_timeout *= 1000000;
       break;
     case '4':
       af = AF_INET;
@@ -455,7 +517,7 @@ void parse_arg (int argc, char **argv)
       fprintf( stderr, "IPv6 not enabled.\n" );
       break;
 #endif
-#ifndef NO_IPINFO
+#ifdef IPINFO
     case 'y':
       ipinfo_no = atoi (optarg);
       if (ipinfo_no < 0)
@@ -464,12 +526,31 @@ void parse_arg (int argc, char **argv)
     case 'z':
       ipinfo_no = 0;
       break;
+#else
+    case 'y':
+    case 'z':
+      fprintf( stderr, "IPINFO not enabled.\n" );
+      break;
+#endif
+#ifdef SO_MARK
+    case 'M':
+      mark = atoi (optarg);
+      if (mark < 0) {
+        fprintf( stderr, "SO_MARK must be positive.\n" );
+        exit(EXIT_FAILURE);
+      }
+      break;
+#else
+    case 'M':
+      fprintf( stderr, "SO_MARK not enabled.\n" );
+      break;
 #endif
     }
   }
 
   if (DisplayMode == DisplayReport ||
       DisplayMode == DisplayTXT ||
+      DisplayMode == DisplayJSON ||
       DisplayMode == DisplayXML ||
       DisplayMode == DisplayRaw ||
       DisplayMode == DisplayCSV)
@@ -540,6 +621,7 @@ int main(int argc, char **argv)
   srand (getpid());
 
   display_detect(&argc, &argv);
+  display_mode = 0;
 
   /* The field options are now in a static array all together,
      but that requires a run-time initialization. */
@@ -566,28 +648,30 @@ int main(int argc, char **argv)
   }
 
   if (PrintHelp) {
-    printf("usage: %s [-hvrwctglspniuT46] [--help] [--version] [--report]\n"
-	   "\t\t[--report-wide] [--report-cycles=COUNT] [--curses] [--gtk]\n"
-           "\t\t[--csv|-C] [--raw] [--split] [--mpls] [--no-dns] [--show-ips]\n"
-           "\t\t[--address interface] [--filename=FILE|-F]\n" /* BL */
-#ifndef NO_IPINFO
-           "\t\t[--ipinfo=item_no|-y item_no]\n"
-           "\t\t[--aslookup|-z]\n"
-#endif
-           "\t\t[--psize=bytes/-s bytes]\n"            /* ok */
-           "\t\t[--report-wide|-w] [-u|-T] [--port=PORT] [--timeout=SECONDS]\n"            /* rew */
-	   "\t\t[--interval=SECONDS] HOSTNAME\n", argv[0]);
+       printf("usage: %s [--help] [--version] [-4|-6] [-F FILENAME]\n"
+              "\t\t[--report] [--report-wide] [--displaymode MODE]\n"
+              "\t\t[--xml] [--gtk] [--curses] [--raw] [--csv] [--json] [--split]\n"
+              "\t\t[--no-dns] [--show-ips] [-o FIELDS] [-y IPINFO] [--aslookup]\n"
+              "\t\t[-i INTERVAL] [-c COUNT] [-s PACKETSIZE] [-B BITPATTERN]\n"
+              "\t\t[-Q TOS] [--mpls]\n"
+              "\t\t[-a ADDRESS] [-f FIRST-TTL] [-m MAX-TTL] [-U MAX_UNKNOWN]\n"
+              "\t\t[--udp] [--tcp] [--sctp] [-P PORT] [-L LOCALPORT] [-Z TIMEOUT]\n"
+              "\t\t[-G GRACEPERIOD] [-M MARK] HOSTNAME\n", argv[0]);
+       printf("See the man page for details.\n");
     exit(0);
   }
 
   time_t now = time(NULL);
+
+  if (!names) append_to_names (argv[0], "localhost"); // default: localhost. 
+
   names_t* head = names;
   while (names != NULL) {
 
     Hostname = names->name;
-    if (Hostname == NULL) Hostname = "localhost";
+    //  if (Hostname == NULL) Hostname = "localhost"; // no longer necessary.
     if (gethostname(LocalHostname, sizeof(LocalHostname))) {
-    strcpy(LocalHostname, "UNKNOWNHOST");
+      strcpy(LocalHostname, "UNKNOWNHOST");
     }
 
     if (net_preopen_result != 0) {
@@ -675,11 +759,11 @@ int main(int argc, char **argv)
       }
     }
 
+
     lock(argv[0], stdout);
       display_open();
       dns_open();
 
-      display_mode = 0;
       display_loop();
 
       net_end_transit();
