@@ -14,12 +14,9 @@
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-    
-   1999-08-13 ok Olav@okvittem.priv.no  added -psize
-
 */
 
-#include <config.h>
+#include "config.h"
 
 #if defined(HAVE_SYS_XTI_H)
 #include <sys/xti.h>
@@ -28,6 +25,8 @@
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/socket.h>
+#include <sys/ioctl.h>
+#include <sys/select.h>
 #include <netinet/in.h>
 #include <memory.h>
 #include <unistd.h>
@@ -64,6 +63,20 @@ struct UDPHeader {
   uint16 checksum;
 };
 
+/* Structure of an TCP header, as far as we need it.  */
+struct TCPHeader {
+  uint16 srcport;
+  uint16 dstport;
+  uint32 seq;
+};
+
+/* Structure of an SCTP header */
+struct SCTPHeader {
+  uint16 srcport;
+  uint16 dstport;
+  uint32 veri_tag;
+};
+
 /* Structure of an IPv4 UDP pseudoheader.  */
 struct UDPv4PHeader {
   uint32 saddr;
@@ -113,7 +126,7 @@ struct nethost {
   int best;
   int worst;
   int avg;	/* average:  addByMin */
-  int gmean;	/* geometirc mean: addByMin */
+  int gmean;	/* geometric mean: addByMin */
   int jitter;	/* current jitter, defined as t1-t0 addByMin */
 /*int jbest;*/	/* min jitter, of cause it is 0, not needed */
   int javg;	/* avg jitter */
@@ -122,6 +135,8 @@ struct nethost {
   int transit;
   int saved[SAVED_PINGS];
   int saved_seq_offset;
+  struct mplslen mpls;
+  struct mplslen mplss[MAXPATH];
 };
 
 
@@ -130,15 +145,10 @@ struct sequence {
   int transit;
   int saved_seq;
   struct timeval time;
+  int socket;
 };
 
 
-/* Configuration parameter: How many queries to unknown hosts do we
-   send? (This limits the amount of traffic generated if a host is not
-   reachable) -- REW */
-#define MAX_UNKNOWN_HOSTS 5
-
-
 /* BSD-derived kernels use host byte order for the IP length and 
    offset fields when using raw sockets.  We detect this automatically at 
    run-time and do the right thing. */
@@ -178,7 +188,7 @@ struct sockaddr_in * rsa4 = (struct sockaddr_in *) &remotesockaddr_struct;
 ip_t * sourceaddress;
 ip_t * remoteaddress;
 
-/* XXX How do I code this to be IPV6 compatible??? -- REW */
+/* XXX How do I code this to be IPV6 compatible??? */
 #ifdef ENABLE_IPV6
 char localaddr[INET6_ADDRSTRLEN];
 #else
@@ -193,12 +203,20 @@ static int numhosts = 10;
 
 extern int fstTTL;		/* initial hub(ttl) to ping byMin */
 extern int maxTTL;		/* last hub to ping byMin*/
+extern int maxUnknown;	/* stop ping threshold */
 extern int cpacketsize;		/* packet size used by ping */
 static int packetsize;		/* packet size used by ping */
+static int spacketsize;		/* packet size used by sendto */
 extern int bitpattern;		/* packet bit pattern used by ping */
 extern int tos;			/* type of service set in ping packet*/
 extern int af;			/* address family of remote target */
 extern int mtrtype;		/* type of query packet used */
+extern int remoteport;          /* target port for TCP tracing */
+extern int localport;  /* source port for UDP tracing */
+extern int tcp_timeout;             /* timeout for TCP connections */
+#ifdef SO_MARK
+extern int mark;		/* SO_MARK to set for ping packet*/
+#endif
 
 /* return the number of microseconds to wait before sending the next
    ping */
@@ -209,32 +227,40 @@ int calc_deltatime (float waittime)
 }
 
 
-/* This doesn't work for odd sz. I don't know enough about this to say
-   that this is wrong. It doesn't seem to cripple mtr though. -- REW */
 int checksum(void *data, int sz) 
 {
-  unsigned short *ch;
-  unsigned int sum;
+  uint16 *ch;
+  uint32 sum;
+  uint16 odd;
 
   sum = 0;
   ch = data;
+  if (sz % 2) {
+    ((char *)&odd)[0] = ((char *)data)[sz - 1];
+    sum = odd;
+  }
   sz = sz / 2;
   while (sz--) {
     sum += *(ch++);
   }
-  
-  sum = (sum >> 16) + (sum & 0xffff);  
+  while (sum >> 16) {
+    sum = (sum >> 16) + (sum & 0xffff);
+  }
 
   return (~sum & 0xffff);  
 }
 
 
 /* Prepend pseudoheader to the udp datagram and calculate checksum */
-int udp_checksum(void *pheader, void *udata, int psize, int dsize)
+int udp_checksum(void *pheader, void *udata, int psize, int dsize, int alt_checksum)
 {
   unsigned int tsize = psize + dsize;
   char csumpacket[tsize];
   memset(csumpacket, (unsigned char) abs(bitpattern), abs(tsize));
+  if (alt_checksum && dsize >= 2) {
+    csumpacket[psize + sizeof(struct UDPHeader)] = 0;
+    csumpacket[psize + sizeof(struct UDPHeader) + 1] = 0;
+  }
 
   struct UDPv4PHeader *prepend = (struct UDPv4PHeader *) csumpacket;
   struct UDPv4PHeader *udppheader = (struct UDPv4PHeader *) pheader;
@@ -255,14 +281,9 @@ int udp_checksum(void *pheader, void *udata, int psize, int dsize)
 }
 
 
-int new_sequence(int index) 
+void save_sequence(int index, int seq)
 {
-  static int next_sequence = MinSequence;
-  int seq;
-
-  seq = next_sequence++;
-  if (next_sequence >= MaxSequence)
-    next_sequence = MinSequence;
+  display_rawxmit(index, seq);
 
   sequence[seq].index = index;
   sequence[seq].transit = 1;
@@ -274,34 +295,295 @@ int new_sequence(int index)
     host[index].up = 0;
   host[index].sent = 1;
   net_save_xmit(index);
-  
+}
+
+int new_sequence(int index)
+{
+  static int next_sequence = MinSequence;
+  int seq;
+
+  seq = next_sequence++;
+  if (next_sequence >= MaxSequence)
+    next_sequence = MinSequence;
+
+  save_sequence(index, seq);
+
   return seq;
 }
 
+/*  Attempt to connect to a TCP port with a TTL */
+void net_send_tcp(int index)
+{
+  int ttl, s;
+  int opt = 1;
+  int port;
+  struct sockaddr_storage local;
+  struct sockaddr_storage remote;
+  struct sockaddr_in *local4 = (struct sockaddr_in *) &local;
+  struct sockaddr_in6 *local6 = (struct sockaddr_in6 *) &local;
+  struct sockaddr_in *remote4 = (struct sockaddr_in *) &remote;
+  struct sockaddr_in6 *remote6 = (struct sockaddr_in6 *) &remote;
+  socklen_t len;
+
+  ttl = index + 1;
+
+  s = socket(af, SOCK_STREAM, 0);
+  if (s < 0) {
+    display_clear();
+    perror("socket()");
+    exit(EXIT_FAILURE);
+  }
+
+  memset(&local, 0, sizeof (local));
+  memset(&remote, 0, sizeof (remote));
+  local.ss_family = af;
+  remote.ss_family = af;
+
+  switch (af) {
+  case AF_INET:
+    addrcpy((void *) &local4->sin_addr, (void *) &ssa4->sin_addr, af);
+    addrcpy((void *) &remote4->sin_addr, (void *) remoteaddress, af);
+    remote4->sin_port = htons(remoteport);
+    len = sizeof (struct sockaddr_in);
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    addrcpy((void *) &local6->sin6_addr, (void *) &ssa6->sin6_addr, af);
+    addrcpy((void *) &remote6->sin6_addr, (void *) remoteaddress, af);
+    remote6->sin6_port = htons(remoteport);
+    len = sizeof (struct sockaddr_in6);
+    break;
+#endif
+  }
+
+  if (bind(s, (struct sockaddr *) &local, len)) {
+    display_clear();
+    perror("bind()");
+    exit(EXIT_FAILURE);
+  }
+
+  if (getsockname(s, (struct sockaddr *) &local, &len)) {
+    display_clear();
+    perror("getsockname()");
+    exit(EXIT_FAILURE);
+  }
+
+  opt = 1;
+  if (ioctl(s, FIONBIO, &opt)) {
+    display_clear();
+    perror("ioctl FIONBIO");
+    exit(EXIT_FAILURE);
+  }
+
+  switch (af) {
+  case AF_INET:
+    if (setsockopt(s, IPPROTO_IP, IP_TTL, &ttl, sizeof (ttl))) {
+      display_clear();
+      perror("setsockopt IP_TTL");
+      exit(EXIT_FAILURE);
+    }
+    if (setsockopt(s, IPPROTO_IP, IP_TOS, &tos, sizeof (tos))) {
+      display_clear();
+      perror("setsockopt IP_TOS");
+      exit(EXIT_FAILURE);
+    }
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    if (setsockopt(s, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &ttl, sizeof (ttl))) {
+      display_clear();
+      perror("setsockopt IP_TTL");
+      exit(EXIT_FAILURE);
+    }
+    break;
+#endif
+  }
+
+#ifdef SO_MARK
+    if (mark >= 0 && setsockopt( s, SOL_SOCKET, SO_MARK, &mark, sizeof mark ) ) {
+      perror( "setsockopt SO_MARK" );
+      exit( EXIT_FAILURE );
+    }
+#endif
+
+  switch (local.ss_family) {
+  case AF_INET:
+    port = ntohs(local4->sin_port);
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    port = ntohs(local6->sin6_port);
+    break;
+#endif
+  default:
+    display_clear();
+    perror("unknown AF?");
+    exit(EXIT_FAILURE);
+  }
+
+  save_sequence(index, port);
+  gettimeofday(&sequence[port].time, NULL);
+  sequence[port].socket = s;
+
+  connect(s, (struct sockaddr *) &remote, len);
+}
+
+/*  Attempt to connect to a SCTP port with a TTL */
+void net_send_sctp(int index)
+{
+  int ttl, s;
+  int opt = 1;
+  int port;
+  struct sockaddr_storage local;
+  struct sockaddr_storage remote;
+  struct sockaddr_in *local4 = (struct sockaddr_in *) &local;
+  struct sockaddr_in6 *local6 = (struct sockaddr_in6 *) &local;
+  struct sockaddr_in *remote4 = (struct sockaddr_in *) &remote;
+  struct sockaddr_in6 *remote6 = (struct sockaddr_in6 *) &remote;
+  socklen_t len;
+
+  ttl = index + 1;
+
+  s = socket(af, SOCK_STREAM, IPPROTO_SCTP);
+  if (s < 0) {
+    display_clear();
+    perror("socket()");
+    exit(EXIT_FAILURE);
+  }
+
+  memset(&local, 0, sizeof (local));
+  memset(&remote, 0, sizeof (remote));
+  local.ss_family = af;
+  remote.ss_family = af;
+
+  switch (af) {
+  case AF_INET:
+    addrcpy((void *) &local4->sin_addr, (void *) &ssa4->sin_addr, af);
+    addrcpy((void *) &remote4->sin_addr, (void *) remoteaddress, af);
+    remote4->sin_port = htons(remoteport);
+    len = sizeof (struct sockaddr_in);
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    addrcpy((void *) &local6->sin6_addr, (void *) &ssa6->sin6_addr, af);
+    addrcpy((void *) &remote6->sin6_addr, (void *) remoteaddress, af);
+    remote6->sin6_port = htons(remoteport);
+    len = sizeof (struct sockaddr_in6);
+    break;
+#endif
+  }
+
+  if (bind(s, (struct sockaddr *) &local, len)) {
+    display_clear();
+    perror("bind()");
+    exit(EXIT_FAILURE);
+  }
+
+  if (getsockname(s, (struct sockaddr *) &local, &len)) {
+    display_clear();
+    perror("getsockname()");
+    exit(EXIT_FAILURE);
+  }
+
+  opt = 1;
+  if (ioctl(s, FIONBIO, &opt)) {
+    display_clear();
+    perror("ioctl FIONBIO");
+    exit(EXIT_FAILURE);
+  }
+
+  switch (af) {
+  case AF_INET:
+    if (setsockopt(s, IPPROTO_IP, IP_TTL, &ttl, sizeof (ttl))) {
+      display_clear();
+      perror("setsockopt IP_TTL");
+      exit(EXIT_FAILURE);
+    }
+    if (setsockopt(s, IPPROTO_IP, IP_TOS, &tos, sizeof (tos))) {
+      display_clear();
+      perror("setsockopt IP_TOS");
+      exit(EXIT_FAILURE);
+    }
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    if (setsockopt(s, IPPROTO_IPV6, IPV6_UNICAST_HOPS, &ttl, sizeof (ttl))) {
+      display_clear();
+      perror("setsockopt IP_TTL");
+      exit(EXIT_FAILURE);
+    }
+    break;
+#endif
+  }
+
+#ifdef SO_MARK
+    if (mark >= 0 && setsockopt( s, SOL_SOCKET, SO_MARK, &mark, sizeof mark ) ) {
+      perror( "setsockopt SO_MARK" );
+      exit( EXIT_FAILURE );
+    }
+#endif
+
+  switch (local.ss_family) {
+  case AF_INET:
+    port = ntohs(local4->sin_port);
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    port = ntohs(local6->sin6_port);
+    break;
+#endif
+  default:
+    display_clear();
+    perror("unknown AF?");
+    exit(EXIT_FAILURE);
+  }
+
+  save_sequence(index, port);
+  gettimeofday(&sequence[port].time, NULL);
+  sequence[port].socket = s;
+
+  connect(s, (struct sockaddr *) &remote, len);
+}
 
 /*  Attempt to find the host at a particular number of hops away  */
 void net_send_query(int index) 
 {
+  if (mtrtype == IPPROTO_TCP) {
+    net_send_tcp(index);
+    return;
+  }
+  
+  if (mtrtype == IPPROTO_SCTP) {
+    net_send_sctp(index);
+    return;
+  }
+
   /*ok  char packet[sizeof(struct IPHeader) + sizeof(struct ICMPHeader)];*/
   char packet[MAXPACKET];
   struct IPHeader *ip = (struct IPHeader *) packet;
   struct ICMPHeader *icmp = NULL;
   struct UDPHeader *udp = NULL;
   struct UDPv4PHeader *udpp = NULL;
+  uint16 checksum_result;
   uint16 mypid;
 
   /*ok  int packetsize = sizeof(struct IPHeader) + sizeof(struct ICMPHeader) + datasize;*/
   int rv;
   static int first=1;
-  int ttl, iphsize = 0, echotype = 0, salen = 0, udphsize = 0;
+  int ttl, iphsize = 0, echotype = 0, salen = 0;
 
   ttl = index + 1;
 
+#ifdef ENABLE_IPV6
   /* offset for ipv6 checksum calculation */
   int offset = 6;
+#endif
 
   if ( packetsize < MINPACKET ) packetsize = MINPACKET;
   if ( packetsize > MAXPACKET ) packetsize = MAXPACKET;
+  if ( mtrtype == IPPROTO_UDP && remoteport && packetsize < (MINPACKET + 2)) {
+    packetsize = MINPACKET + 2;
+  }
 
   memset(packet, (unsigned char) abs(bitpattern), abs(packetsize));
 
@@ -350,6 +632,13 @@ void net_send_query(int index)
 #endif
   }
 
+#ifdef SO_MARK
+    if (mark >= 0 && setsockopt( sendsock, SOL_SOCKET, SO_MARK, &mark, sizeof mark ) ) {
+      perror( "setsockopt SO_MARK" );
+      exit( EXIT_FAILURE );
+    }
+#endif
+
   switch ( mtrtype ) {
   case IPPROTO_ICMP:
     icmp = (struct ICMPHeader *)(packet + iphsize);
@@ -365,20 +654,28 @@ void net_send_query(int index)
 
   case IPPROTO_UDP:
     udp = (struct UDPHeader *)(packet + iphsize);
-    udphsize = sizeof (struct UDPHeader);
     udp->checksum  = 0;
-    mypid = (uint16)getpid();
-    if (mypid < MinPort)
-      mypid += MinPort;
-
+    if (!localport) {
+      mypid = (uint16)getpid();
+      if (mypid < MinPort)
+        mypid += MinPort;
+    } else {
+      mypid = (uint16)localport;
+    }
     udp->srcport = htons(mypid);
-    udp->length = abs(packetsize) - iphsize;
-    if(!BSDfix)
-      udp->length = htons(udp->length);
- 
-    udp->dstport = new_sequence(index);
-    gettimeofday(&sequence[udp->dstport].time, NULL);
-    udp->dstport = htons(udp->dstport);
+    udp->length = htons(abs(packetsize) - iphsize);
+
+    if (!remoteport) {
+      udp->dstport = new_sequence(index);
+      gettimeofday(&sequence[udp->dstport].time, NULL);
+      udp->dstport = htons(udp->dstport);
+    } else {
+      // keep dstport constant, stuff sequence into the checksum
+      udp->dstport = htons(remoteport);
+      udp->checksum = new_sequence(index);
+      gettimeofday(&sequence[udp->checksum].time, NULL);
+      udp->checksum = htons(udp->checksum);
+    }
     break;
   }
 
@@ -387,13 +684,22 @@ void net_send_query(int index)
     switch ( mtrtype ) {
     case IPPROTO_UDP:
       /* checksum is not mandatory. only calculate if we know ip->saddr */
-      if (ip->saddr) {
+      if (udp->checksum) {
         udpp = (struct UDPv4PHeader *)(malloc(sizeof(struct UDPv4PHeader)));
         udpp->saddr = ip->saddr;
         udpp->daddr = ip->daddr;
         udpp->protocol = ip->protocol;
         udpp->len = udp->length;
-        udp->checksum = udp_checksum(udpp, udp, sizeof(struct UDPv4PHeader), abs(packetsize) - iphsize);
+        checksum_result = udp_checksum(udpp, udp, sizeof(struct UDPv4PHeader), abs(packetsize) - iphsize, 1);
+        packet[iphsize + sizeof(struct UDPHeader)] = ((char *)&checksum_result)[0];
+        packet[iphsize + sizeof(struct UDPHeader) + 1] = ((char *)&checksum_result)[1];
+      } else if (ip->saddr) {
+        udpp = (struct UDPv4PHeader *)(malloc(sizeof(struct UDPv4PHeader)));
+        udpp->saddr = ip->saddr;
+        udpp->daddr = ip->daddr;
+        udpp->protocol = ip->protocol;
+        udpp->len = udp->length;
+        udp->checksum = udp_checksum(udpp, udp, sizeof(struct UDPv4PHeader), abs(packetsize) - iphsize, 0);
       }
       break;
     }
@@ -405,6 +711,9 @@ void net_send_query(int index)
     switch ( mtrtype ) {
     case IPPROTO_UDP:
       /* kernel checksum calculation */
+      if (udp->checksum) {
+        offset = sizeof(struct UDPHeader);
+      }
       if ( setsockopt(sendsock, IPPROTO_IPV6, IPV6_CHECKSUM, &offset, sizeof(offset)) ) {
         perror( "setsockopt IPV6_CHECKSUM" );
         exit( EXIT_FAILURE);
@@ -415,13 +724,15 @@ void net_send_query(int index)
 #endif
   }
 
-  rv = sendto(sendsock, packet, abs(packetsize), 0, 
-	      remotesockaddr, salen);
+  /* sendto() assumes packet length includes the IPv4 header but not the 
+     IPv6 header. */
+  spacketsize = abs(packetsize)	-
+		( ( af == AF_INET ) ? 0 : sizeof (struct ip6_hdr) );
+  rv = sendto(sendsock, packet, spacketsize, 0, remotesockaddr, salen);
   if (first && (rv < 0) && ((errno == EINVAL) || (errno == EMSGSIZE))) {
     /* Try the first packet again using host byte order. */
-    ip->len = abs (packetsize);
-    rv = sendto(sendsock, packet, abs(packetsize), 0, 
-		remotesockaddr, salen);
+    ip->len = spacketsize;
+    rv = sendto(sendsock, packet, spacketsize, 0, remotesockaddr, salen);
     if (rv >= 0) {
       BSDfix = 1;
     }
@@ -432,7 +743,7 @@ void net_send_query(int index)
 
 /*   We got a return on something we sent out.  Record the address and
      time.  */
-void net_process_ping(int seq, void * addr, struct timeval now) 
+void net_process_ping(int seq, struct mplslen mpls, void * addr, struct timeval now) 
 {
   int index;
   int totusec;
@@ -455,6 +766,11 @@ void net_process_ping(int seq, void * addr, struct timeval now)
     return;
   sequence[seq].transit = 0;
 
+  if (sequence[seq].socket > 0) {
+    close(sequence[seq].socket);
+    sequence[seq].socket = 0;
+  }
+
   index = sequence[seq].index;
 
   totusec = (now.tv_sec  - sequence[seq].time.tv_sec ) * 1000000 +
@@ -465,10 +781,12 @@ void net_process_ping(int seq, void * addr, struct timeval now)
 		(void *) &unspec_addr, af ) == 0 ) {
     /* should be out of if as addr can change */
     addrcpy( (void *) &(host[index].addr), addrcopy, af );
+    host[index].mpls = mpls;
     display_rawhost(index, (void *) &(host[index].addr));
 
-  /* multi paths by Min */
+  /* multi paths */
     addrcpy( (void *) &(host[index].addrs[0]), addrcopy, af );
+    host[index].mplss[0] = mpls;
   } else {
     for( i=0; i<MAXPATH; ) {
       if( addrcmp( (void *) &(host[index].addrs[i]), (void *) &addrcopy,
@@ -480,11 +798,9 @@ void net_process_ping(int seq, void * addr, struct timeval now)
     if( addrcmp( (void *) &(host[index].addrs[i]), addrcopy, af ) != 0 && 
         i<MAXPATH ) {
       addrcpy( (void *) &(host[index].addrs[i]), addrcopy, af );
-      
-      /* rafaelmartins: multi path support to '--raw' */
+      host[index].mplss[i] = mpls;
       display_rawhost(index, (void *) &(host[index].addrs[i]));
     }
-  /* end multi paths */
   }
 
   host[index].jitter = totusec - host[index].last;
@@ -502,7 +818,6 @@ void net_process_ping(int seq, void * addr, struct timeval now)
    * at least in linux 2.4.x.
    *  safe guard 1) best[index]>=best[index-1] if index>0
    *             2) best >= average-20,000 usec (good number?)
-   *  Min
   if (index > 0) {
     if (totusec < host[index].best &&
        totusec>= host[index-1].best) host[index].best  = totusec;
@@ -517,10 +832,9 @@ void net_process_ping(int seq, void * addr, struct timeval now)
 	host[index].jworst = host[index].jitter;
 
   host[index].returned++;
-  /* begin addByMin do more stats */
   oldavg = host[index].avg;
   host[index].avg += (totusec - oldavg +.0) / host[index].returned;
-  host[index].var += (totusec - oldavg +.0) * (totusec - host[index].avg);
+  host[index].var += (totusec - oldavg +.0) * (totusec - host[index].avg) / 1000000;
 
   oldjavg = host[index].javg;
   host[index].javg += (host[index].jitter - oldjavg) / host[index].returned;
@@ -530,13 +844,12 @@ void net_process_ping(int seq, void * addr, struct timeval now)
   if ( host[index].returned > 1 )
   host[index].gmean = pow( (double) host[index].gmean, (host[index].returned-1.0)/host[index].returned )
 			* pow( (double) totusec, 1.0/host[index].returned );
-  /* end addByMin*/
   host[index].sent = 0;
   host[index].up = 1;
   host[index].transit = 0;
 
   net_save_return(index, sequence[seq].saved_seq, totusec);
-  display_rawping(index, totusec);
+  display_rawping(index, totusec, seq);
 }
 
 
@@ -558,11 +871,17 @@ void net_process_return(void)
   int num;
   struct ICMPHeader *header = NULL;
   struct UDPHeader *udpheader = NULL;
+  struct TCPHeader *tcpheader = NULL;
+  struct SCTPHeader *sctpheader = NULL;
   struct timeval now;
   ip_t * fromaddress = NULL;
   int echoreplytype = 0, timeexceededtype = 0, unreachabletype = 0;
   int sequence = 0;
 
+  /* MPLS decoding */
+  struct mplslen mpls;
+  mpls.labels = 0;
+
   gettimeofday(&now, NULL);
   switch ( af ) {
   case AF_INET:
@@ -621,6 +940,10 @@ void net_process_return(void)
         header = (struct ICMPHeader *)(packet + sizeof (struct IPHeader) + 
                                                 sizeof (struct ICMPHeader) + 
                                                 sizeof (struct IPHeader));
+
+        if(num > 160)
+          decodempls(num, packet, &mpls, 156);
+
       break;
 #ifdef ENABLE_IPV6
       case AF_INET6:
@@ -630,6 +953,10 @@ void net_process_return(void)
         header = (struct ICMPHeader *) ( packet + 
                                          sizeof (struct ICMPHeader) +
                                          sizeof (struct ip6_hdr) );
+
+        if(num > 140)
+          decodempls(num, packet, &mpls, 136);
+
         break;
 #endif
       }
@@ -654,6 +981,10 @@ void net_process_return(void)
         udpheader = (struct UDPHeader *)(packet + sizeof (struct IPHeader) +
                                                   sizeof (struct ICMPHeader) +
                                                   sizeof (struct IPHeader));
+
+        if(num > 160)
+          decodempls(num, packet, &mpls, 156);
+
       break;
 #ifdef ENABLE_IPV6
       case AF_INET6:
@@ -663,16 +994,97 @@ void net_process_return(void)
         udpheader = (struct UDPHeader *) ( packet +
                                            sizeof (struct ICMPHeader) +
                                            sizeof (struct ip6_hdr) );
+
+        if(num > 140)
+          decodempls(num, packet, &mpls, 136);
+
         break;
 #endif
       }
-      sequence = ntohs(udpheader->dstport);
+      if (remoteport && remoteport == ntohs(udpheader->dstport)) {
+        sequence = ntohs(udpheader->checksum);
+      } else if (!remoteport) {
+        sequence = ntohs(udpheader->dstport);
+      }
+    }
+    break;
+
+  case IPPROTO_TCP:
+    if (header->type == timeexceededtype || header->type == unreachabletype) {
+      switch ( af ) {
+      case AF_INET:
+
+        if ((size_t) num < sizeof(struct IPHeader) +
+                           sizeof(struct ICMPHeader) +
+                           sizeof (struct IPHeader) +
+                           sizeof (struct TCPHeader))
+          return;
+        tcpheader = (struct TCPHeader *)(packet + sizeof (struct IPHeader) +
+                                                  sizeof (struct ICMPHeader) +
+                                                  sizeof (struct IPHeader));
+
+        if(num > 160)
+          decodempls(num, packet, &mpls, 156);
+
+      break;
+#ifdef ENABLE_IPV6
+      case AF_INET6:
+        if ( num < sizeof (struct ICMPHeader) +
+                   sizeof (struct ip6_hdr) + sizeof (struct TCPHeader) )
+          return;
+        tcpheader = (struct TCPHeader *) ( packet +
+                                           sizeof (struct ICMPHeader) +
+                                           sizeof (struct ip6_hdr) );
+
+        if(num > 140)
+          decodempls(num, packet, &mpls, 136);
+
+        break;
+#endif
+      }
+      sequence = ntohs(tcpheader->srcport);
+    }
+    break;
+    
+  case IPPROTO_SCTP:
+    if (header->type == timeexceededtype || header->type == unreachabletype) {
+      switch ( af ) {
+      case AF_INET:
+
+        if ((size_t) num < sizeof(struct IPHeader) +
+                           sizeof(struct ICMPHeader) +
+                           sizeof (struct IPHeader) +
+                           sizeof (struct SCTPHeader))
+          return;
+        sctpheader = (struct SCTPHeader *)(packet + sizeof (struct IPHeader) +
+                                                  sizeof (struct ICMPHeader) +
+                                                  sizeof (struct IPHeader));
+
+        if(num > 160)
+          decodempls(num, packet, &mpls, 156);
+
+      break;
+#ifdef ENABLE_IPV6
+      case AF_INET6:
+        if ( num < sizeof (struct ICMPHeader) +
+                   sizeof (struct ip6_hdr) + sizeof (struct SCTPHeader) )
+          return;
+        sctpheader = (struct SCTPHeader *) ( packet +
+                                           sizeof (struct ICMPHeader) +
+                                           sizeof (struct ip6_hdr) );
+
+        if(num > 140)
+          decodempls(num, packet, &mpls, 136);
+
+        break;
+#endif
+      }
+      sequence = ntohs(sctpheader->srcport);
     }
     break;
   }
-
   if (sequence)
-    net_process_ping(sequence, (void *)fromaddress, now);
+    net_process_ping (sequence, mpls, (void *) fromaddress, now);
 }
 
 
@@ -687,6 +1099,15 @@ ip_t *net_addrs(int at, int i)
   return (ip_t *)&(host[at].addrs[i]);
 }
 
+void *net_mpls(int at)
+{
+  return (struct mplslen *)&(host[at].mplss);
+}
+
+void *net_mplss(int at, int i)
+{
+  return (struct mplslen *)&(host[at].mplss[i]);
+}
 
 int net_loss(int at) 
 {
@@ -736,14 +1157,13 @@ int net_gmean(int at)
 int net_stdev(int at) 
 {
   if( host[at].returned > 1 ) {
-    return ( sqrt( host[at].var/(host[at].returned -1.0) ) );
+    return ( 1000.0 * sqrt( host[at].var/(host[at].returned -1.0) ) );
   } else {
     return( 0 );
   }
 }
 
 
-/* jitter stuff */
 int net_jitter(int at) 
 { 
   return (host[at].jitter); 
@@ -766,7 +1186,6 @@ int net_jinta(int at)
 { 
   return (host[at].jinta); 
 }
-/* end jitter */
 
 
 int net_max(void) 
@@ -775,8 +1194,7 @@ int net_max(void)
   int max;
 
   max = 0;
-  /* replacedByMin
-     for(at = 0; at < MaxHost-2; at++) { */
+  /* for(at = 0; at < MaxHost-2; at++) { */
   for(at = 0; at < maxTTL-1; at++) {
     if ( addrcmp( (void *) &(host[at].addr),
                   (void *) remoteaddress, af ) == 0 ) {
@@ -791,14 +1209,12 @@ int net_max(void)
 }
 
 
-/* add by Min (wonder its named net_min;-)) because of ttl stuff */
 int net_min (void) 
 {
   return ( fstTTL - 1 );
 }
 
 
-/* Added by Brian Casey December 1997 bcasey@imagiware.com*/
 int net_returned(int at) 
 { 
   return host[at].returned;
@@ -855,7 +1271,7 @@ int net_send_batch(void)
            As our detination range (in the example 10000) is much 
            smaller (reasonable packet sizes), and our rand() range much 
            larger, this effect is insignificant. Oh! That other formula
-           didn't work. -- REW */
+           didn't work. */
       packetsize = MINPACKET + rand () % (-cpacketsize - MINPACKET);
     } else {
       packetsize = cpacketsize;
@@ -876,8 +1292,8 @@ int net_send_batch(void)
     /* The second condition in the next "if" statement was added in mtr-0.56, 
 	but I don't remember why. It makes mtr stop skipping sections of unknown
 	hosts. Removed in 0.65. 
-	If the line proves neccesary, it should at least NOT trigger that line 
-	when host[i].addr == 0 -- REW */
+	If the line proves necessary, it should at least NOT trigger that line
+	when host[i].addr == 0 */
     if ( ( addrcmp( (void *) &(host[i].addr),
                     (void *) remoteaddress, af ) == 0 )
 	/* || (host[i].addr == host[batch_at].addr)  */)
@@ -887,8 +1303,8 @@ int net_send_batch(void)
   if (	/* success in reaching target */
      ( addrcmp( (void *) &(host[batch_at].addr),
                 (void *) remoteaddress, af ) == 0 ) ||
-      /* fail in consecuitive MAX_UNKNOWN_HOSTS (firewall?) */
-      (n_unknown > MAX_UNKNOWN_HOSTS) ||
+      /* fail in consecutive maxUnknown (firewall?) */
+      (n_unknown > maxUnknown) ||
       /* or reach limit  */
       (batch_at >= maxTTL-1)) {
     numhosts = batch_at+1;
@@ -1102,16 +1518,93 @@ void net_reset(void)
   
   for (at = 0; at < MaxSequence; at++) {
     sequence[at].transit = 0;
+    if (sequence[at].socket > 0) {
+      close(sequence[at].socket);
+      sequence[at].socket = 0;
+    }
   }
 
   gettimeofday(&reset, NULL);
 }
 
+int net_set_interfaceaddress_udp()
+{
+#ifdef ENABLE_IPV6
+  struct sockaddr_storage name_struct;
+#else
+  struct sockaddr_in name_struct;
+#endif
+  struct sockaddr_in *  sa4;
+  struct sockaddr_in6 * sa6;
+  struct sockaddr * name = (struct sockaddr *) &name_struct;
+  struct sockaddr_storage remote;
+  struct sockaddr_in *remote4 = (struct sockaddr_in *) &remote;
+  struct sockaddr_in6 *remote6 = (struct sockaddr_in6 *) &remote;
+  socklen_t len;
+  int s;
+
+  memset(&remote, 0, sizeof (remote));
+  remote.ss_family = af;
+
+  switch (af) {
+  case AF_INET:
+    addrcpy((void *) &remote4->sin_addr, (void *) remoteaddress, af);
+    remote4->sin_port = htons(remoteport);
+    len = sizeof (struct sockaddr_in);
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    addrcpy((void *) &remote6->sin6_addr, (void *) remoteaddress, af);
+    remote6->sin6_port = htons(remoteport);
+    len = sizeof (struct sockaddr_in6);
+    break;
+#endif
+  }
+
+  s = socket (af, SOCK_DGRAM, 0);
+  if (s < 0) {
+    perror("udp socket()");
+    exit(EXIT_FAILURE);
+  }
+
+  if (connect(s, (struct sockaddr *) &remote, len)) {
+    perror("udp connect() failed");
+    exit(EXIT_FAILURE);
+  }
+
+  getsockname(s, name, &len);
+  sockaddrtop( name, localaddr, sizeof localaddr );
+  switch (af) {
+  case AF_INET:
+    sa4 = (struct sockaddr_in *) name;
+    addrcpy((void*)&ssa4->sin_addr, (void *) &(sa4->sin_addr), af );
+    break;
+#ifdef ENABLE_IPV6
+  case AF_INET6:
+    sa6 = (struct sockaddr_in6 *) name;
+    addrcpy((void*)&ssa6->sin6_addr, (void *) &(sa6->sin6_addr), af );
+    break;
+#endif
+  }
+  close(s);
+
+  return 0;
+}
+
 
 int net_set_interfaceaddress (char *InterfaceAddress)
 {
-  int len = 0;
+#ifdef ENABLE_IPV6
+  struct sockaddr_storage name_struct;
+#else
+  struct sockaddr_in name_struct;
+#endif
+  struct sockaddr * name = (struct sockaddr *) &name_struct;
+  socklen_t len = 0;
 
+  if (mtrtype == IPPROTO_UDP && remoteport && !InterfaceAddress) {
+    return net_set_interfaceaddress_udp();
+  }
   if (!InterfaceAddress) return 0; 
 
   sourcesockaddr->sa_family = af;
@@ -1140,6 +1633,8 @@ int net_set_interfaceaddress (char *InterfaceAddress)
     perror("mtr: failed to bind to interface");
       return( 1 );
   }
+  getsockname (sendsock, name, &len);
+  sockaddrtop( name, localaddr, sizeof localaddr );
   return 0; 
 }
 
@@ -1195,7 +1690,7 @@ void net_save_return(int at, int seq, int ms)
 {
   int idx;
   idx = seq - host[at].saved_seq_offset;
-  if (idx < 0 || idx > SAVED_PINGS) {
+  if (idx < 0 || idx >= SAVED_PINGS) {
     return;
   }
   host[at].saved[idx] = ms;
@@ -1211,7 +1706,7 @@ void sockaddrtop( struct sockaddr * saddr, char * strptr, size_t len ) {
   switch ( saddr->sa_family ) {
   case AF_INET:
     sa4 = (struct sockaddr_in *) saddr;
-    strncpy( strptr, inet_ntoa( (struct in_addr) sa4->sin_addr ), len - 1 );
+    strncpy( strptr, inet_ntoa( sa4->sin_addr ), len - 1 );
     strptr[ len - 1 ] = '\0';
     return;
 #ifdef ENABLE_IPV6
@@ -1259,3 +1754,109 @@ void addrcpy( char * a, char * b, int af ) {
 #endif
   }
 }
+
+/* Decode MPLS */
+void decodempls(int num, char *packet, struct mplslen *mpls, int offset) {
+
+  int i;
+  unsigned int ext_ver, ext_res, ext_chk, obj_hdr_len;
+  u_char obj_hdr_class, obj_hdr_type;
+
+  /* loosely derived from the traceroute-nanog.c
+   * decoding by Jorge Boncompte */
+  ext_ver = packet[offset]>>4;
+  ext_res = (packet[offset]&15)+ packet[offset+1];
+  ext_chk = ((unsigned int)packet[offset+2]<<8)+packet[offset+3];
+
+  /* Check for ICMP extension header */
+  if (ext_ver == 2 && ext_res == 0 && ext_chk != 0 && num >= (offset+6)) {
+    obj_hdr_len = ((int)packet[offset+4]<<8)+packet[offset+5];
+    obj_hdr_class = packet[offset+6];
+    obj_hdr_type = packet[offset+7];
+
+    /* make sure we have an MPLS extension */
+    if (obj_hdr_len >= 8 && obj_hdr_class == 1 && obj_hdr_type == 1) {
+      /* how many labels do we have?  will be at least 1 */
+      mpls->labels = (obj_hdr_len-4)/4;
+
+      /* save all label objects */
+      for(i=0; (i<mpls->labels) && (i < MAXLABELS) && (num >= (offset+8)+(i*4)); i++) {
+
+        /* piece together the 20 byte label value */
+        mpls->label[i] = ((unsigned long) (packet[(offset+8)+(i*4)] << 12 & 0xff000) +
+            (unsigned int) (packet[(offset+9)+(i*4)] << 4 & 0xff0) +
+            (packet[(offset+10)+(i*4)] >> 4 & 0xf));
+        mpls->exp[i] = (packet[(offset+10)+(i*4)] >> 1) & 0x7;
+        mpls->s[i] = (packet[(offset+10)+(i*4)] & 0x1); /* should be 1 if only one label */
+        mpls->ttl[i] = packet[(offset+11)+(i*4)];
+      }
+    }
+  }
+}
+
+/* Add open sockets to select() */
+void net_add_fds(fd_set *writefd, int *maxfd)
+{
+  int at, fd;
+  for (at = 0; at < MaxSequence; at++) {
+    fd = sequence[at].socket;
+    if (fd > 0) {
+      FD_SET(fd, writefd);
+      if (fd >= *maxfd)
+        *maxfd = fd + 1;
+    }
+  }
+}
+
+/* check if we got connection or error on any fds */
+void net_process_fds(fd_set *writefd)
+{
+  int at, fd, r;
+  struct timeval now;
+  uint64_t unow, utime;
+
+  /* Can't do MPLS decoding */
+  struct mplslen mpls;
+  mpls.labels = 0;
+
+  gettimeofday(&now, NULL);
+  unow = now.tv_sec * 1000000L + now.tv_usec;
+
+  for (at = 0; at < MaxSequence; at++) {
+    fd = sequence[at].socket;
+    if (fd > 0 && FD_ISSET(fd, writefd)) {
+      r = write(fd, "G", 1);
+      /* if write was successful, or connection refused we have
+       * (probably) reached the remote address. Anything else happens to the
+       * connection, we write it off to avoid leaking sockets */
+      if (r == 1 || errno == ECONNREFUSED)
+        net_process_ping(at, mpls, remoteaddress, now);
+      else if (errno != EAGAIN) {
+        close(fd);
+        sequence[at].socket = 0;
+      }
+    }
+    if (fd > 0) {
+      utime = sequence[at].time.tv_sec * 1000000L + sequence[at].time.tv_usec;
+      if (unow - utime > tcp_timeout) {
+        close(fd);
+        sequence[at].socket = 0;
+      }
+    }
+  }
+}
+
+/* for GTK frontend */
+void net_harvest_fds(void)
+{
+  fd_set writefd;
+  int maxfd = 0;
+  struct timeval tv;
+
+  FD_ZERO(&writefd);
+  tv.tv_sec = 0;
+  tv.tv_usec = 0;
+  net_add_fds(&writefd, &maxfd);
+  select(maxfd, NULL, &writefd, NULL, &tv);
+  net_process_fds(&writefd);
+}
