@@ -3,9 +3,8 @@
     Copyright (C) 1997,1998  Matt Kimball
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
+    it under the terms of the GNU General Public License version 2 as 
+    published by the Free Software Foundation.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -23,7 +22,9 @@ void report_open(void);
 void report_close(void);
 void txt_open(void);
 void txt_close(void);
+void json_open(void);
+void json_close(void);
 void xml_open(void);
 void xml_close(void);
 void csv_open(void);
-void csv_close(void);
+void csv_close(time_t now);
