@@ -16,7 +16,8 @@
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
-#include <config.h>
+#include "config.h"
+
 #include <sys/types.h>
 #include <sys/time.h>
 #include <stdlib.h>
@@ -29,30 +30,42 @@
 #include <errno.h>
 
 #include "mtr.h"
-#include "display.h"
 #include "dns.h"
 #include "net.h"
+#include "asn.h"
+#include "display.h"
 
 extern int Interactive;
 extern int MaxPing;
 extern int ForceMaxPing;
 extern float WaitTime;
+extern float GraceTime;
 double dnsinterval;
+extern int mtrtype;
 
 static struct timeval intervaltime;
 int display_offset = 0;
 
+#define GRACETIME (GraceTime * 1000*1000)
 
 void select_loop(void) {
   fd_set readfd;
+  fd_set writefd;
   int anyset = 0;
   int maxfd = 0;
   int dnsfd, netfd;
+#ifdef ENABLE_IPV6
+  int dnsfd6;
+#endif
   int NumPing = 0;
   int paused = 0;
   struct timeval lasttime, thistime, selecttime;
+  struct timeval startgrace;
   int dt;
   int rv; 
+  int graceperiod = 0;
+
+  memset(&startgrace, 0, sizeof(startgrace));
 
   gettimeofday(&lasttime, NULL);
 
@@ -62,6 +75,7 @@ void select_loop(void) {
     intervaltime.tv_usec = dt % 1000000;
 
     FD_ZERO(&readfd);
+    FD_ZERO(&writefd);
 
     maxfd = 0;
 
@@ -70,6 +84,18 @@ void select_loop(void) {
       maxfd = 1;
     }
 
+#ifdef ENABLE_IPV6
+    if (dns) {
+      dnsfd6 = dns_waitfd6();
+      if (dnsfd6 >= 0) {
+        FD_SET(dnsfd6, &readfd);
+        if(dnsfd6 >= maxfd) maxfd = dnsfd6 + 1;
+      } else {
+        dnsfd6 = 0;
+      }
+    } else
+      dnsfd6 = 0;
+#endif
     if (dns) {
       dnsfd = dns_waitfd();
       FD_SET(dnsfd, &readfd);
@@ -81,12 +107,20 @@ void select_loop(void) {
     FD_SET(netfd, &readfd);
     if(netfd >= maxfd) maxfd = netfd + 1;
 
+    if (mtrtype == IPPROTO_TCP)
+      net_add_fds(&writefd, &maxfd);
+
     do {
       if(anyset || paused) {
+	/* Set timeout to 0.1s.
+	 * While this is almost instantaneous for human operators,
+	 * it's slow enough for computers to go do something else;
+	 * this prevents mtr from hogging 100% CPU time on one core.
+	 */
 	selecttime.tv_sec = 0;
-	selecttime.tv_usec = 0;
+	selecttime.tv_usec = paused?100000:0; 
       
-	rv = select(maxfd, (void *)&readfd, NULL, NULL, &selecttime);
+	rv = select(maxfd, (void *)&readfd, &writefd, NULL, &selecttime);
 
       } else {
 	if(Interactive) display_redraw();
@@ -97,10 +131,24 @@ void select_loop(void) {
 	   (thistime.tv_sec == lasttime.tv_sec + intervaltime.tv_sec &&
 	    thistime.tv_usec >= lasttime.tv_usec + intervaltime.tv_usec)) {
 	  lasttime = thistime;
-	  if(NumPing >= MaxPing && (!Interactive || ForceMaxPing))
+
+	  if (!graceperiod) {
+	    if (NumPing >= MaxPing && (!Interactive || ForceMaxPing)) {
+	      graceperiod = 1;
+	      startgrace = thistime;
+	    }
+
+	    /* do not send out batch when we've already initiated grace period */
+	    if (!graceperiod && net_send_batch())
+	      NumPing++;
+	  }
+	}
+
+	if (graceperiod) {
+	  dt = (thistime.tv_usec - startgrace.tv_usec) +
+		    1000000 * (thistime.tv_sec - startgrace.tv_sec);
+	  if (dt > GRACETIME)
 	    return;
-	  if (net_send_batch())
-	    NumPing++;
 	}
 
 	selecttime.tv_usec = (thistime.tv_usec - lasttime.tv_usec);
@@ -148,7 +196,13 @@ void select_loop(void) {
     }
 
     /*  Have we finished a nameservice lookup?  */
-    if(dns && FD_ISSET(dnsfd, &readfd)) {
+#ifdef ENABLE_IPV6
+    if(dns && dnsfd6 && FD_ISSET(dnsfd6, &readfd)) {
+      dns_ack6();
+      anyset = 1;
+    }
+#endif
+    if(dns && dnsfd && FD_ISSET(dnsfd, &readfd)) {
       dns_ack();
       anyset = 1;
     }
@@ -184,6 +238,16 @@ void select_loop(void) {
 	  display_clear();
 	}
 	break;
+#ifdef IPINFO
+      case ActionII:
+	ipinfo_no++;
+	if (ipinfo_no > ipinfo_max)
+	  ipinfo_no = 0;
+	break;
+      case ActionAS:
+	ipinfo_no = ipinfo_no?0:ipinfo_max;
+	break;
+#endif
 
       case ActionScrollDown:
         display_offset += 5;
@@ -197,6 +261,10 @@ void select_loop(void) {
       }
       anyset = 1;
     }
+
+    /* Check for activity on open sockets */
+    if (mtrtype == IPPROTO_TCP)
+      net_process_fds(&writefd);
   }
   return;
 }
